#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include "obdnet.h"
#include "resource.h"

/* Function Prototypes */
LRESULT CALLBACK OBDNetDlgProc(HWND hWndDlg, UINT Message, WPARAM wParam, LPARAM lParam);

/* DLL Handle*/
HMODULE hOBDNetModule = NULL;

/* Library Handle */
OBDNET_HANDLE hOBDNetHandle;

/* Dialog Handle (Needed for PID List Callback */
HWND hDialog;

/* Outputs Debug Information */
void Debug(HWND hDlg, char szMsg[])
{
	SendDlgItemMessage(hDlg, IDC_DEBUG, EM_SETSEL, -1, -1);
	SendDlgItemMessage(hDlg, IDC_DEBUG, EM_REPLACESEL, NULL, (LPARAM)szMsg);
	SendDlgItemMessage(hDlg, IDC_DEBUG, EM_REPLACESEL, NULL, (LPARAM)"\r\n");
}

/* Callback for PID List */
void OBDNET_CALLBACK addToPidList(OBDNET_HANDLE OBDNetHandle, const char * szUniqueId)
{
	char szTemp[255];

	sprintf(szTemp, "PID List Callback: \"%s\"", szUniqueId);
	Debug(hDialog, szTemp);

	SendDlgItemMessage(hDialog, IDC_PIDUNIQUEID, CB_ADDSTRING, NULL, (LPARAM)szUniqueId);
}

/* Callback for Selected PID List */
void OBDNET_CALLBACK addToSelectedList(OBDNET_HANDLE OBDNetHandle, const char * szUniqueId)
{
	char szTemp[255];

	sprintf(szTemp, "Selected PID List Callback: \"%s\"", szUniqueId);
	Debug(hDialog, szTemp);

	SendDlgItemMessage(hDialog, IDC_SELECTEDLIST, LB_ADDSTRING, NULL, (LPARAM)szUniqueId);
}

/* Callback for PID Streaming */
void OBDNET_CALLBACK streamPidCallback(OBDNET_HANDLE OBDNetHandle, const OBDNET_PID_VALUE * lpPidValue)
{
	static int iFirst = GetTickCount();
	static int iCnt = 0;

	iCnt++;
	char szTemp[255];

	sprintf(szTemp, "streamPidCallback()\r\nUnique ID: %s\r\nTimeStamp: %u", lpPidValue->UniqueId, lpPidValue->TimeStamp);
	Debug(hDialog, szTemp);

	if(lpPidValue->PidValueType == OBDNET_PID_VALUETYPE_NUMBER)
	{
		sprintf(szTemp, "English Value:\t%g", lpPidValue->EnglishValue);
		Debug(hDialog, szTemp);

		sprintf(szTemp, "Metric Value:\t%g", lpPidValue->MetricValue);
		Debug(hDialog, szTemp);
	}
	else if(lpPidValue->PidValueType == OBDNET_PID_VALUETYPE_STRING)
	{
		sprintf(szTemp, "English Value:\t%s", lpPidValue->EnglishValueString);
		Debug(hDialog, szTemp);
		
		sprintf(szTemp, "Metric Value:\t%s", lpPidValue->MetricValueString);
		Debug(hDialog, szTemp);
	}

	int iLast = GetTickCount();

	sprintf(szTemp, "%i - FPS: %g", iCnt, iCnt / ((iLast - iFirst) / 1000.0));
	Debug(hDialog, szTemp);
}

/* Application Entry Point */
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
//	INITCOMMONCONTROLSEX controls;
//	controls.dwSize = sizeof(INITCOMMONCONTROLSEX);
//	controls.dwICC = 0x00004000 | ICC_WIN95_CLASSES | 256 | 512;
//	InitCommonControlsEx(&controls);

	LoadLibrary("RichEd32.dll");

//	int iReturn = DialogBox(hInstance, MAKEINTRESOURCE(IDD_OBDNET_TEST_DIALOG), NULL, (DLGPROC)OBDNetDlgProc);
	HWND hDlg = CreateDialog(hInstance, MAKEINTRESOURCE(IDD_OBDNET_TEST_DIALOG), NULL, (DLGPROC)OBDNetDlgProc);

	if(hDlg == NULL)
	{
		char szTemp[255];

		sprintf(szTemp, "CreateDialog() Failed!\nGetLastError(): %i", GetLastError());
		MessageBox(NULL, szTemp, "Error!", MB_OK);
		return 0;
	}

	ShowWindow(hDlg, SW_SHOW);

	bool bRun = true;
	MSG Msg;

	while(bRun)
	{
		if(PeekMessage(&Msg, NULL, NULL, NULL, PM_REMOVE))
		{
			if(Msg.message == WM_QUIT)
			{
				bRun = false;
			}
			else
			{
				TranslateMessage(&Msg);
				DispatchMessage(&Msg);
			}
		}
		else
		{
			if(hOBDNetHandle != NULL)
			{
				OBDNetWorker(hOBDNetHandle);
			}
			Sleep(1);
		}
	}

	return 0;
}

/* Dialog Procedure */
LRESULT CALLBACK OBDNetDlgProc(HWND hWndDlg, UINT Message, WPARAM wParam, LPARAM lParam)
{
	switch(Message)
	{
	case WM_INITDIALOG:
		{
			// Used for PID List Callback
			hDialog = hWndDlg;

			/* Default Address and Port */
			SetDlgItemText(hWndDlg, IDC_IPADDRESS, "127.0.0.1");
			SetDlgItemText(hWndDlg, IDC_PORTNUMBER, "1301");

			/* Add Logging Modes */
			SendDlgItemMessage(hWndDlg, IDC_LOGGINGMODE, CB_ADDSTRING, NULL, (LPARAM)"Unknown");
			SendDlgItemMessage(hWndDlg, IDC_LOGGINGMODE, CB_ADDSTRING, NULL, (LPARAM)"Stopped");
			SendDlgItemMessage(hWndDlg, IDC_LOGGINGMODE, CB_ADDSTRING, NULL, (LPARAM)"Monitoring");
			SendDlgItemMessage(hWndDlg, IDC_LOGGINGMODE, CB_ADDSTRING, NULL, (LPARAM)"Recording");
			SendDlgItemMessage(hWndDlg, IDC_LOGGINGMODE, CB_SETCURSEL, 0, NULL);
		}
		return TRUE;
		break;

	case WM_COMMAND:
		switch(wParam)
		{
			case IDCANCEL:
			case IDOK:
				EndDialog(hWndDlg, 0);
				PostQuitMessage(0);		// Modeless Dialog
				break;

			case IDC_INITIALIZE:
				{
					char szTemp[1024];

					OBDNET_STATUS iInit = OBDNetInitialize(OBDNET_APITYPE_C_CPLUSPLUS, NULL);

					sprintf(szTemp, "OBDNetInitialize(): %i\r\nDescription: %s", iInit, OBDNetStatusGetErrorCodeString(iInit));
					Debug(hWndDlg, szTemp);

					hOBDNetHandle = OBDNetCreate();
					
					sprintf(szTemp, "OBDNetCreate(): %08X", hOBDNetHandle);
					Debug(hWndDlg, szTemp);

					/* Setup Callbacks */
					OBDNET_STATUS iPidList = OBDNetSetCallback(OBDNET_CALLBACK_PID_LIST_ITEM, addToPidList);
					OBDNET_STATUS iSelList = OBDNetSetCallback(OBDNET_CALLBACK_SELECTED_PID_ITEM, addToSelectedList);
					OBDNET_STATUS iStreamPid = OBDNetSetCallback(OBDNET_CALLBACK_STREAMING_PID_VALUE, streamPidCallback);
					
					sprintf(szTemp, "OBDNetSetCallback(OBDNET_CALLBACK_PID_LIST_ITEM): %i\r\nDescription: %s", iPidList, OBDNetStatusGetErrorCodeString(iPidList));
					Debug(hWndDlg, szTemp);
					
					sprintf(szTemp, "OBDNetSetCallback(OBDNET_CALLBACK_SELECTED_PID_ITEM): %i\r\nDescription: %s", iSelList, OBDNetStatusGetErrorCodeString(iSelList));
					Debug(hWndDlg, szTemp);
					
					sprintf(szTemp, "OBDNetSetCallback(OBDNET_CALLBACK_STREAMING_PID_VALUE): %i\r\nDescription: %s", iStreamPid, OBDNetStatusGetErrorCodeString(iStreamPid));
					Debug(hWndDlg, szTemp);
				}
				break;

			case IDC_CONNECT:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Connecting!");
						break;
					}
					
					char szIpAddress[255];
					char szPort[255];
					GetDlgItemText(hWndDlg, IDC_IPADDRESS, szIpAddress, 255);
					GetDlgItemText(hWndDlg, IDC_PORTNUMBER, szPort, 255);
					int iPort = atoi(szPort);

					Debug(hWndDlg, "Opening Connection...");
					OBDNET_STATUS connection = OBDNetOpenConnection(hOBDNetHandle, szIpAddress, iPort, 30000);

					char szTemp[1024];

					sprintf(szTemp, "OBDNetOpenConnection(): %i", connection);
					Debug(hWndDlg, szTemp);

					sprintf(szTemp, "Description: %s", OBDNetStatusGetErrorCodeString(connection));
					Debug(hWndDlg, szTemp);
				}
				break;

			case IDC_GETPIDLIST:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Retrieving PID List!");
						break;
					}

					char szTemp[255];
					
					Debug(hWndDlg, "Attempting to Get PID List");

					OBDNET_STATUS iPidList = OBDNetParameterGetPidList(hOBDNetHandle);

					sprintf(szTemp, "OBDNetParameterGetPidList(): %i\r\nDescription: %s", iPidList, OBDNetStatusGetErrorCodeString(iPidList));
					Debug(hWndDlg, szTemp);
				}
				break;

			case IDC_GETPIDCOUNT:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Retrieving PID Count!");
						break;
					}


					char szTemp[255];
					
					unsigned int iPids = OBDNetParameterGetCount(hOBDNetHandle);

					sprintf(szTemp, "OBDNetParameterGetCount(): %i", iPids);
					Debug(hWndDlg, szTemp);

					SetDlgItemInt(hWndDlg, IDC_PIDCOUNT, iPids, FALSE);
				}
				break;

			case IDC_GETPIDVALUE:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Retrieving PID Value!");
						break;
					}

					char szUniqueId[255];
					int iIndex;

					iIndex = SendDlgItemMessage(hWndDlg, IDC_PIDUNIQUEID, CB_GETCURSEL, NULL, NULL);
					SendDlgItemMessage(hWndDlg, IDC_PIDUNIQUEID, CB_GETLBTEXT, iIndex, (LPARAM)szUniqueId);

					if(iIndex == CB_ERR)
						Debug(hWndDlg, "Select a PID First!");
					else
					{
						OBDNET_PID_INFO pidInfo;
						OBDNET_PID_VALUE pidValue;

						OBDNET_STATUS iGetInfo = OBDNetParameterGetInfo(hOBDNetHandle, szUniqueId, &pidInfo);
						OBDNET_STATUS iScanValue = OBDNetParameterGetCurrentValue(hOBDNetHandle, szUniqueId, &pidValue);

						char szTemp[255];

						{
							DWORD	Time1;
							DWORD	Time2;
							DWORD	TimeDifference;
							DWORD	CallCount = 5000;
							DWORD	CallIndex;

							Time1 = GetTickCount();

							for (CallIndex=0; CallIndex<CallCount; CallIndex++)
							{
								iScanValue = OBDNetParameterGetCurrentValue(hOBDNetHandle, szUniqueId, &pidValue);
							}

							Time2 = GetTickCount();
							TimeDifference = Time2 - Time1;

							sprintf(szTemp, "%u calls made in %ums!", CallCount, TimeDifference);
							MessageBox(NULL, szTemp, "Benchmark Results", MB_OK);
						}

						sprintf(szTemp, "OBDNetParameterGetInfo(): %i\r\nDescription: %s", iGetInfo, OBDNetStatusGetErrorCodeString(iGetInfo));
						Debug(hWndDlg, szTemp);

						sprintf(szTemp, "OBDNetParameterGetCurrentValue(): %i\r\nDescription: %s", iScanValue, OBDNetStatusGetErrorCodeString(iScanValue));
						Debug(hWndDlg, szTemp);

						SetDlgItemText(hWndDlg, IDC_UNITSENGLISH, pidInfo.EnglishUnits);
						SetDlgItemText(hWndDlg, IDC_UNITSMETRIC, pidInfo.MetricUnits);

						if(iScanValue == OBDNET_STATUS_SUCCESS)
						{
							// Success, Fill Boxes with Data
							if(pidValue.PidValueType == OBDNET_PID_VALUETYPE_STRING)
							{
								// Strings
								SetDlgItemText(hWndDlg, IDC_VALUEENGLISH, pidValue.EnglishValueString);
								SetDlgItemText(hWndDlg, IDC_VALUEMETRIC, pidValue.MetricValueString);
							}
							else
							{
								// Numbers
								char szTemp[255];

								sprintf(szTemp, "%g", pidValue.EnglishValue);
								SetDlgItemText(hWndDlg, IDC_VALUEENGLISH, szTemp);

								sprintf(szTemp, "%g", pidValue.MetricValue);
								SetDlgItemText(hWndDlg, IDC_VALUEMETRIC, szTemp);
							}
						}
						else
						{
							// Not Successful, Clear Boxes of Data
							SetDlgItemText(hWndDlg, IDC_VALUEENGLISH, "");
							SetDlgItemText(hWndDlg, IDC_VALUEMETRIC, "");
						}
					}
				}
				break;

			case IDC_GETSELECTEDLIST:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Retrieving the Selected List!");
						break;
					}

					// Clear List Box
					SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_RESETCONTENT, NULL, NULL);

					OBDNET_STATUS iCallback = OBDNetSetCallback(OBDNET_CALLBACK_SELECTED_PID_ITEM, addToSelectedList);
					OBDNET_STATUS iGetList = OBDNetLoggingGetSelectedPidList(hOBDNetHandle);

					char szTemp[255];

					sprintf(szTemp, "OBDNetSetCallback(): %i\r\nDescription: %s", iCallback, OBDNetStatusGetErrorCodeString(iCallback));
					Debug(hWndDlg, szTemp);

					sprintf(szTemp, "OBDNetLoggingGetSelectedPidList(): %i\r\nDescription: %s", iGetList, OBDNetStatusGetErrorCodeString(iGetList));
					Debug(hWndDlg, szTemp);
				}
				break;

			case IDC_CLEARSELECTEDLIST:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Clearing the Selected List!");
						break;
					}

					OBDNET_STATUS iClear = OBDNetLoggingClearSelectedPidList(hOBDNetHandle);

					if(iClear == OBDNET_STATUS_SUCCESS)
						SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_RESETCONTENT, NULL, NULL);

					char szTemp[255];

					sprintf(szTemp, "OBDNetLoggingClearSelectedPidList(): %i\r\nDescription: \"%s\"", iClear, OBDNetStatusGetErrorCodeString(iClear));
				}
				break;

			case IDC_ADDPID:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Adding a PID to the Selected List!");
						break;
					}
					
					char szUniqueId[255];
					int iIndex;
					
					iIndex = SendDlgItemMessage(hWndDlg, IDC_PIDUNIQUEID, CB_GETCURSEL, NULL, NULL);
					SendDlgItemMessage(hWndDlg, IDC_PIDUNIQUEID, CB_GETLBTEXT, iIndex, (LPARAM)szUniqueId);
					
					if(iIndex == CB_ERR)
						Debug(hWndDlg, "Select a PID First!");
					else
					{
						// Add the PID
						OBDNET_STATUS iAdd = OBDNetLoggingAddSelectedPid(hOBDNetHandle, szUniqueId);
						if(iAdd == OBDNET_STATUS_SUCCESS)
						{
							if(SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_FINDSTRINGEXACT, -1, (LPARAM)szUniqueId) != LB_ERR)
								SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_ADDSTRING, NULL, (LPARAM)szUniqueId);
						}

						char szTemp[255];

						sprintf(szTemp, "OBDNetLoggingAddSelectedPid(): %i\r\nDescription: \"%s\"", iAdd, OBDNetStatusGetErrorCodeString(iAdd));
						Debug(hWndDlg, szTemp);
					}
				}
				break;

			case IDC_REMOVEPID:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Removing a PID from the Selected List!");
						break;
					}
					
					char szUniqueId[255];
					int iIndex;
					
					iIndex = SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_GETCURSEL, NULL, NULL);
					SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_GETTEXT, iIndex, (LPARAM)szUniqueId);
					
					if(iIndex == CB_ERR)
						Debug(hWndDlg, "Select a PID First!");
					else
					{
						// Remove the PID
						OBDNET_STATUS iRemove = OBDNetLoggingRemoveSelectedPid(hOBDNetHandle, szUniqueId);
						
						if(iRemove == OBDNET_STATUS_SUCCESS)
							SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_DELETESTRING, iIndex, NULL);

						char szTemp[255];
						
						sprintf(szTemp, "OBDNetLoggingRemoveSelectedPid(): %i\r\nDescription: \"%s\"", iRemove, OBDNetStatusGetErrorCodeString(iRemove));
						Debug(hWndDlg, szTemp);
					}
				}
				break;

			case IDC_GETLOGGINGMODE:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Retrieving the Logging Mode!");
						break;
					}

					int iGetMode = OBDNetLoggingGetMode(hOBDNetHandle);

					char szTemp[255];

					sprintf(szTemp, "OBDNetLoggingGetMode(): %i", iGetMode);
					Debug(hWndDlg, szTemp);

					SendDlgItemMessage(hWndDlg, IDC_LOGGINGMODE, CB_SETCURSEL, iGetMode, NULL);

					switch(iGetMode)
					{
					case OBDNET_LOGGINGMODE_UNKNOWN:
						Debug(hWndDlg, "Current Logging Mode: Unknown");
						break;

					case OBDNET_LOGGINGMODE_STOPPED:
						Debug(hWndDlg, "Current Logging Mode: Stopped");
						break;

					case OBDNET_LOGGINGMODE_MONITORING:
						Debug(hWndDlg, "Current Logging Mode: Monitoring");
						break;

					case OBDNET_LOGGINGMODE_RECORDING:
						Debug(hWndDlg, "Current Logging Mode: Recording");
						break;
					}
				}
				break;

			case IDC_SETLOGGINGMODE:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Setting the Logging Mode!");
						break;
					}

					int iIndex;
					OBDNET_STATUS iMode;
					
					iIndex = SendDlgItemMessage(hWndDlg, IDC_LOGGINGMODE, CB_GETCURSEL, NULL, NULL);

					if(iIndex != CB_ERR)
						iMode = OBDNetLoggingSetMode(hOBDNetHandle, iIndex);

					char szTemp[255];

					sprintf(szTemp, "OBDNetLoggingSetMode(): %i\r\nDescription: %s", iMode, OBDNetStatusGetErrorCodeString(iMode));
					Debug(hWndDlg, szTemp);
				}
				break;

			case IDC_STARTSTREAMINGPID:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Streaming a PID!");
						break;
					}
					
					char szUniqueId[255];
					int iIndex;
					
					iIndex = SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_GETCURSEL, NULL, NULL);
					SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_GETTEXT, iIndex, (LPARAM)szUniqueId);
					
					if(iIndex == CB_ERR)
						Debug(hWndDlg, "Get the Selected List First!\r\nSelect a PID First!");
					else
					{
						// Steam the PID
						OBDNET_STATUS iStart = OBDNetLoggingStartStreamingPid(hOBDNetHandle, szUniqueId);
						
						char szTemp[255];
						
						sprintf(szTemp, "OBDNetLoggingStartStreamingPid(): %i\r\nDescription: \"%s\"", iStart, OBDNetStatusGetErrorCodeString(iStart));
						Debug(hWndDlg, szTemp);
					}
				}
				break;

			case IDC_STOPSTREAMINGPID:
				{
					if(hOBDNetHandle == NULL)
					{
						Debug(hWndDlg, "Must Initialize OBDNet Before Streaming a PID!");
						break;
					}
					
					char szUniqueId[255];
					int iIndex;
					
					iIndex = SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_GETCURSEL, NULL, NULL);
					SendDlgItemMessage(hWndDlg, IDC_SELECTEDLIST, LB_GETTEXT, iIndex, (LPARAM)szUniqueId);
					
					if(iIndex == CB_ERR)
						Debug(hWndDlg, "Get the Selected List First!\r\nSelect a PID First!");
					else
					{
						// Stop Streaming PID
						OBDNET_STATUS iStart = OBDNetLoggingStopStreamingPid(hOBDNetHandle, szUniqueId);
						
						char szTemp[255];
						
						sprintf(szTemp, "OBDNetLoggingStopStreamingPid(): %i\r\nDescription: \"%s\"", iStart, OBDNetStatusGetErrorCodeString(iStart));
						Debug(hWndDlg, szTemp);
					}
				}
				break;
		}
		break;
	}

	return FALSE;
}